@@ -1,3 +1,5 @@
+import sys
+
 import asyncio
 import json
 import logging
@@ -10,12 +12,15 @@ from zmq_tubes.matcher import TopicMatcher
 
 
 class TubeException(Exception): pass            # flake8: E701
-class TubeTopicNotConfigured(Exception): pass   # flake8: E701
-class TubeMessageError(Exception): pass         # flake8: E701
-class TubeMessageTimeout(Exception): pass       # flake8: E701
-class TubeMethodNotSupported(Exception): pass         # flake8: E701
+class TubeTopicNotConfigured(TubeException): pass   # flake8: E701
+class TubeMessageError(TubeException): pass         # flake8: E701
+class TubeMessageTimeout(TubeException): pass       # flake8: E701
+class TubeMethodNotSupported(TubeException): pass   # flake8: E701
+class TubeConnectionError(TubeException): pass   # flake8: E701
 
 
+LESS38 = sys.version_info < (3, 8)
+
 TUBE_TYPE_MAPPING = {
     'SUB': zmq.SUB,
     'PUB': zmq.PUB,
@@ -339,43 +344,62 @@ class Tube:
         """
         raw_msg = message.format_message()
         self.logger.debug("Send (tube: %s) to %s", self.name, raw_msg)
+        if not message.raw_socket or message.raw_socket.closed:
+            raise TubeConnectionError(
+                f'The tube {message.tube.name} is already closed.')
         try:
             message.raw_socket.send_multipart(raw_msg)
         except (TypeError, zmq.ZMQError) as ex:
             raise TubeMessageError(
                 f"The message '{message}' does not be sent.") from ex
 
-    async def request(self, *args, **kwargs) -> TubeMessage:
+    async def request(self, *args, post_send_callback=None,
+                      **kwargs) -> TubeMessage:
         """
         Send request
         :param request: Optional[TubeMessage]
         :param topic: Optional[str]
         :param payload: Optional[dict]
         :param timeout: int
+        :param post_send_callback: Optional[Callable]
         :return:
         """
         if args:
             if isinstance(args[0], TubeMessage):
-                return await self.__request_message(*args, **kwargs)
+                return await self.__request_message(
+                    *args, post_send_callback=post_send_callback, **kwargs
+                )
             elif isinstance(args[0], str):
-                return await self.__request_payload(*args, **kwargs)
+                return await self.__request_payload(
+                    *args, post_send_callback=post_send_callback, **kwargs
+                )
         elif kwargs:
             if 'message' in kwargs:
-                return await self.__request_message(**kwargs)
+                return await self.__request_message(
+                    post_send_callback=post_send_callback,
+                    **kwargs
+                )
             elif 'topic' in kwargs:
-                return await self.__request_payload(**kwargs)
+                return await self.__request_payload(
+                    post_send_callback=post_send_callback,
+                    **kwargs
+                )
         raise NotImplementedError("Unknown type of topic")
 
-    async def __request_payload(self, topic: str, payload=None, timeout=None):
+    async def __request_payload(self, topic: str, payload=None, timeout=None,
+                                post_send_callback=None):
         request = TubeMessage(
             self,
             payload=payload,
             topic=topic,
             raw_socket=self.raw_socket
         )
-        return await self.__request_message(request, timeout)
+        return await self.__request_message(
+            request, timeout=timeout, post_send_callback=post_send_callback
+        )
 
-    async def __request_message(self, request: TubeMessage, timeout: int = 30):
+    async def __request_message(self, request: TubeMessage, timeout: int = 30,
+                                post_send_callback=None):
         if self.tube_type != zmq.REQ:
             raise TubeMethodNotSupported(
                 f"The tube '{self.name}' (type: '{self.tube_type_name}') "
@@ -383,7 +407,12 @@ class Tube:
             )
         try:
             self.send(request)
-            while await request.raw_socket.poll(timeout * 1000) != 0:
+            if post_send_callback:
+                if asyncio.iscoroutinefunction(post_send_callback):
+                    await post_send_callback(request)
+                else:
+                    post_send_callback(request)
+            if await request.raw_socket.poll(timeout * 1000) != 0:
                 response = await self.receive_data(
                     raw_socket=request.raw_socket)
                 if response.topic != request.topic:
@@ -391,10 +420,15 @@ class Tube:
                         f"The response comes to different topic "
                         f"({request.topic} != {response.topic}).")
                 return response
+            else:
+                self.logger.error("The request timout")
         finally:
             if not self.is_persistent:
                 self.logger.debug(f"Close tube {self.name}")
-                request.raw_socket.close()
+                if request.raw_socket and not request.raw_socket.closed:
+                    request.raw_socket.close()
+        if self.is_closed:
+            raise TubeConnectionError(f'The tube {self.name} was closed.')
         raise TubeMessageTimeout(
             f"No answer for the request in {timeout}s. Topic: {request.topic}")
 
@@ -424,7 +458,9 @@ class TubeNode:
 
     def __enter__(self):
         self.connect()
-        self.start()
+        args = {} if LESS38 else {'name': 'zmq/main'}
+        asyncio.create_task(self.start(), **args)
+        return self
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         self.stop()
@@ -530,13 +566,14 @@ class TubeNode:
                                              f'dealer.')
         tube.send(topic, payload)
 
-    async def request(self, topic: str, payload=None, timeout=30) \
-            -> TubeMessage:
+    async def request(self, topic: str, payload=None, timeout=30,
+                      post_send_callback=None) -> TubeMessage:
         tube = self.get_tube_by_topic(topic, [zmq.REQ])
         if not tube:
             raise TubeTopicNotConfigured(f'The topic "{topic}" is not assigned '
                                          f'to any Tube for request.')
-        res = await tube.request(topic, payload, timeout)
+        res = await tube.request(topic, payload, timeout=timeout,
+                                 post_send_callback=post_send_callback)
         return res
 
     def publish(self, topic: str, payload=None):
@@ -609,6 +646,7 @@ class TubeNode:
         self.logger.info("The main loop was started.")
         while not self._stop_main_loop:
             events = await poller.poll(timeout=100)
+            # print(events)
             for event in events:
                 raw_socket = event[0]
                 tube: Tube = raw_socket.__dict__['tube']
@@ -622,17 +660,21 @@ class TubeNode:
                         )
                     continue
                 if tube.tube_type == zmq.SUB:
+                    args = {} if LESS38 else {'name': 'zmq/sub'}
                     for callback in callbacks:
-                        loop.create_task(callback(request), name='zmq/sub')
+                        loop.create_task(callback(request), **args)
                 elif tube.tube_type == zmq.REP:
+                    args = {} if LESS38 else {'name': 'zmq/rep'}
                     loop.create_task(
                         _callback_wrapper(callbacks[-1], request),
-                        name='zmq/rep')
+                        **args)
                 elif tube.tube_type == zmq.ROUTER:
+                    args = {} if LESS38 else {'name': 'zmq/router'}
                     loop.create_task(
                         _callback_wrapper(callbacks[-1], request),
-                        name='zmq/router'
+                        **args
                     )
                 elif tube.tube_type == zmq.DEALER:
-                    loop.create_task(callbacks[-1](request), name='zmq/dealer')
+                    args = {} if LESS38 else {'name': 'zmq/dealer'}
+                    loop.create_task(callbacks[-1](request), **args)
         self.logger.info("The main loop was ended.")
