@@ -1,20 +1,19 @@
-from functools import singledispatchmethod
 from threading import Thread, Lock, Event, current_thread
 import zmq
 from zmq import Poller, Context
 
 from .manager import TubeMessage, Tube as AsyncTube, TubeNode as AsyncTubeNode,\
     TubeMethodNotSupported, TubeMessageError, TubeMessageTimeout, \
-    TubeTopicNotConfigured
+    TubeTopicNotConfigured, TubeConnectionError
 
 
 class TubeThreadDeadLock(Exception): pass
 
 
-class DaemonThread(Thread):
+class StoppableThread(Thread):
     def __init__(self, *args, **kwargs):
         self.stop_event = Event()
-        kwargs['daemon'] = True
+        # kwargs['daemon'] = True
         super().__init__(*args, **kwargs)
 
     def is_stopped(self):
@@ -31,34 +30,16 @@ class Tube(AsyncTube):
         self.lock = Lock()
         self.context = Context().instance()
 
-    @singledispatchmethod
-    def send(self, arg):
-        raise NotImplementedError("Unknown type of topic")
-
-    @send.register
-    def _(self, topic: str, payload=None, raw_socket=None):
-        """
-        Send payload to topic.
-        :param topic - topic
-        :param payload - payload
-        :param raw_socket - zmqSocket, it used for non permanent connection
-        """
-        message = TubeMessage(
-            self,
-            payload=payload,
-            topic=topic,
-            raw_socket=raw_socket if raw_socket else self.raw_socket
-        )
-        self.send(message)
-
-    @send.register
-    def _(self, message: TubeMessage):
+    def __send_message(self, message: TubeMessage):
         """
         Send message.
         :param message - TubeMessage
         """
         raw_msg = message.format_message()
         self.logger.debug("Send (tube: %s) to %s", self.name, raw_msg)
+        if not message.raw_socket or message.raw_socket.closed:
+            raise TubeConnectionError(
+                f'The tube {message.tube.name} is already closed.')
         if not self.lock.acquire(timeout=10):
             raise TubeThreadDeadLock()
         try:
@@ -69,22 +50,49 @@ class Tube(AsyncTube):
         finally:
             self.lock.release()
 
-    @singledispatchmethod
-    def request(self, arg) -> TubeMessage:
+    def request(self, *args, post_send_callback=None, **kwargs) -> TubeMessage:
+        """
+                Send request
+                :param request: Optional[TubeMessage]
+                :param topic: Optional[str]
+                :param payload: Optional[dict]
+                :param timeout: int
+                :param post_send_callback: Optional[Callable]
+                :return: TubeMessage
+                """
+        if args:
+            if isinstance(args[0], TubeMessage):
+                return self.__request_message(
+                    *args, post_send_callback=post_send_callback, **kwargs
+                )
+            elif isinstance(args[0], str):
+                return self.__request_payload(
+                    *args, post_send_callback=post_send_callback, **kwargs
+                )
+        elif kwargs:
+            if 'message' in kwargs:
+                return self.__request_message(
+                    post_send_callback=post_send_callback, **kwargs
+                )
+            elif 'topic' in kwargs:
+                return self.__request_payload(
+                    post_send_callback=post_send_callback, **kwargs
+                )
         raise NotImplementedError("Unknown type of topic")
 
-    @request.register
-    def _(self, topic: str, payload=None, timeout=None):
+    def __request_payload(self, topic: str, payload=None, timeout=None,
+                          post_send_callback=None):
         request = TubeMessage(
             self,
             payload=payload,
             topic=topic,
-            raw_socket=self.raw_socket
+            raw_socket=self.raw_socket,
         )
-        return self.request(request, timeout)
+        return self.request(request, timeout=timeout,
+                            post_send_callback=post_send_callback)
 
-    @request.register
-    def _(self, request: TubeMessage, timeout: int = 30):
+    def __request_message(self, request: TubeMessage, timeout: int = 30,
+                          post_send_callback=None):
         if self.tube_type != zmq.REQ:
             raise TubeMethodNotSupported(
                 f"The tube '{self.name}' (type: '{self.tube_type_name}') "
@@ -92,25 +100,24 @@ class Tube(AsyncTube):
             )
         try:
             self.send(request)
-            # if request.raw_socket.poll(timeout * 1000) == zmq.POLLIN:
-            counter = timeout
-            while (res := request.raw_socket.poll(1000)) != 0 or \
-                    (counter != 0 and not self.is_closed):
-                if res != 0:
-                    response = self.receive_data(raw_socket=request.raw_socket)
-                    if response.topic != request.topic:
-                        raise TubeMessageError(
-                            f"The response comes to different topic "
-                            f"({request.topic} != {response.topic}).")
-                    return response
-                elif counter == 0:
-                    self.logger.error("The request timout")
-                    break
-                counter -= 1
+            if post_send_callback:
+                post_send_callback(request)
+            if request.raw_socket.poll(timeout * 1000) != 0:
+                response = self.receive_data(raw_socket=request.raw_socket)
+                if response.topic != request.topic:
+                    raise TubeMessageError(
+                        f"The response comes to different topic "
+                        f"({request.topic} != {response.topic}).")
+                return response
+            else:
+                self.logger.error("The request timout")
         finally:
             if not self.is_persistent:
                 self.logger.debug(f"Close tube {self.name}")
-                request.raw_socket.close(3000)
+                if request.raw_socket and not request.raw_socket.closed:
+                    request.raw_socket.close(1000)
+        if self.is_closed:
+            raise TubeConnectionError(f'The tube {self.name} was closed.')
         raise TubeMessageTimeout(
             f"No answer for the request in {timeout}s. Topic: {request.topic}")
 
@@ -137,13 +144,22 @@ class TubeNode(AsyncTubeNode):
         super().__init__(*args, **kwargs)
         self.__main_thread = None
 
-    def request(self, topic: str, payload=None, timeout=30) \
-            -> TubeMessage:
+    def __enter__(self):
+        self.connect()
+        self.start()
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.stop()
+        self.close()
+
+    def request(self, topic: str, payload=None, timeout=30,
+                post_send_callback=None) -> TubeMessage:
         tube = self.get_tube_by_topic(topic, [zmq.REQ])
         if not tube:
             raise TubeTopicNotConfigured(f'The topic "{topic}" is not assigned '
                                          f'to any Tube for request.')
-        res = tube.request(topic, payload, timeout)
+        res = tube.request(topic, payload, timeout=timeout,
+                           post_send_callback=post_send_callback)
         return res
 
     def stop(self):
@@ -170,6 +186,10 @@ class TubeNode(AsyncTubeNode):
                         "from request object.")
             try:
                 tube.send(response)
+            except TubeConnectionError:
+                self.logger.warning(
+                    f"The client (tube '{tube.name}') closes the socket for "
+                    f"sending answer. Probably timeout.")
             except TubeThreadDeadLock:
                 self.logger.error(
                     f"The tube '{tube.name}' waits more then "
@@ -233,6 +253,6 @@ class TubeNode(AsyncTubeNode):
                             f"3s for access to socket.")
             self.logger.info("The main process was ended.")
 
-        self.__main_thread = DaemonThread(target=_main_loop, name='zmq/main')
+        self.__main_thread = StoppableThread(target=_main_loop, name='zmq/main')
         self.__main_thread.start()
         return self.__main_thread
